@@ -1,5 +1,5 @@
 /* UI_FILE - a generic STDIO like output stream.
-   Copyright (C) 1999-2021 Free Software Foundation, Inc.
+   Copyright (C) 1999-2024 Free Software Foundation, Inc.
 
    This file is part of GDB.
 
@@ -16,8 +16,8 @@
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
 
-#ifndef UI_FILE_H
-#define UI_FILE_H
+#ifndef GDB_UI_FILE_H
+#define GDB_UI_FILE_H
 
 #include <string>
 #include "ui-style.h"
@@ -30,18 +30,30 @@ public:
   ui_file ();
   virtual ~ui_file () = 0;
 
+  ui_file (ui_file &&other) = default;
+
   /* Public non-virtual API.  */
 
   void printf (const char *, ...) ATTRIBUTE_PRINTF (2, 3);
 
-  /* Print a string whose delimiter is QUOTER.  Note that these
-     routines should only be called for printing things which are
-     independent of the language of the program being debugged.  */
+  /* Print a NUL-terminated string whose delimiter is QUOTER.  Note
+     that these routines should only be called for printing things
+     which are independent of the language of the program being
+     debugged.
+
+     This will normally escape backslashes and instances of QUOTER.
+     If QUOTER is 0, it won't escape backslashes or any quoting
+     character.  As a side effect, if you pass the backslash character
+     as the QUOTER, this will escape backslashes as usual, but not any
+     other quoting character.  */
   void putstr (const char *str, int quoter);
 
-  void putstrn (const char *str, int n, int quoter);
+  /* Like putstr, but only print the first N characters of STR.  If
+     ASYNC_SAFE is true, then the output is done via the
+     write_async_safe method.  */
+  void putstrn (const char *str, int n, int quoter, bool async_safe = false);
 
-  int putc (int c);
+  void putc (int c);
 
   void vprintf (const char *, va_list) ATTRIBUTE_PRINTF (2, 0);
 
@@ -70,10 +82,10 @@ public:
   virtual bool isatty ()
   { return false; }
 
-  /* true indicates terminal output behaviour such as cli_styling.
+  /* true indicates terminal output behavior such as cli_styling.
      This default implementation indicates to do terminal output
-     behaviour if the UI_FILE is a tty.  A derived class can override
-     TERM_OUT to have cli_styling behaviour without being a tty.  */
+     behavior if the UI_FILE is a tty.  A derived class can override
+     TERM_OUT to have cli_styling behavior without being a tty.  */
   virtual bool term_out ()
   { return isatty (); }
 
@@ -83,6 +95,56 @@ public:
 
   virtual void flush ()
   {}
+
+  /* If this object has an underlying file descriptor, then return it.
+     Otherwise, return -1.  */
+  virtual int fd () const
+  { return -1; }
+
+  /* Indicate that if the next sequence of characters overflows the
+     line, a newline should be inserted here rather than when it hits
+     the end.  If INDENT is non-zero, it is a number of spaces to be
+     printed to indent the wrapped part on the next line.
+
+     If the line is already overfull, we immediately print a newline and
+     the indentation, and disable further wrapping.
+
+     If we don't know the width of lines, but we know the page height,
+     we must not wrap words, but should still keep track of newlines
+     that were explicitly printed.
+
+     This routine is guaranteed to force out any output which has been
+     squirreled away in the wrap_buffer, so wrap_here (0) can be
+     used to force out output from the wrap_buffer.  */
+  virtual void wrap_here (int indent)
+  {
+  }
+
+  /* Emit an ANSI style escape for STYLE.  */
+  virtual void emit_style_escape (const ui_file_style &style);
+
+  /* Rest the current output style to the empty style.  */
+  virtual void reset_style ();
+
+  /* Print STR, bypassing any paging that might be done by this
+     ui_file.  Note that nearly no code should call this -- it's
+     intended for use by gdb_printf, but nothing else.  */
+  virtual void puts_unfiltered (const char *str)
+  {
+    this->puts (str);
+  }
+
+protected:
+
+  /* The currently applied style.  */
+  ui_file_style m_applied_style;
+
+private:
+
+  /* Helper function for putstr and putstrn.  Print the character C on
+     this stream as part of the contents of a literal string whose
+     delimiter is QUOTER.  */
+  void printchar (int c, int quoter, bool async_safe);
 };
 
 typedef std::unique_ptr<ui_file> ui_file_up;
@@ -109,14 +171,16 @@ class string_file : public ui_file
 {
 public:
   /* Construct a string_file to collect 'raw' output, i.e. without
-     'terminal' behaviour such as cli_styling.  */
+     'terminal' behavior such as cli_styling.  */
   string_file () : m_term_out (false) {};
-  /* If TERM_OUT, construct a string_file with terminal output behaviour
+  /* If TERM_OUT, construct a string_file with terminal output behavior
      such as cli_styling)
      else collect 'raw' output like the previous constructor.  */
   explicit string_file (bool term_out) : m_term_out (term_out) {};
   ~string_file () override;
 
+  string_file (string_file &&other) = default;
+
   /* Override ui_file methods.  */
 
   void write (const char *buf, long length_buf) override;
@@ -130,17 +194,27 @@ public:
   /* string_file-specific public API.  */
 
   /* Accesses the std::string containing the entire output collected
-     so far.
+     so far.  */
+  const std::string &string () { return m_string; }
 
-     Returns a non-const reference so that it's easy to move the
-     string contents out of the string_file.  E.g.:
+  /* Return an std::string containing the entire output collected so far.
 
-      string_file buf;
-      buf.printf (....);
-      buf.printf (....);
-      return std::move (buf.string ());
-  */
-  std::string &string () { return m_string; }
+     The internal buffer is cleared, such that it's ready to build a new
+     string.  */
+  std::string release ()
+  {
+    std::string ret = std::move (m_string);
+    m_string.clear ();
+    return ret;
+  }
+
+  /* Set the internal buffer contents to STR.  Any existing contents are
+     discarded.  */
+  string_file &operator= (std::string &&str)
+  {
+    m_string = std::move (str);
+    return *this;
+  }
 
   /* Provide a few convenience methods with the same API as the
      underlying std::string.  */
@@ -195,6 +269,10 @@ public:
 
   bool can_emit_style_escape () override;
 
+  /* Return the underlying file descriptor.  */
+  int fd () const override
+  { return m_fd; }
+
   FILE *get_stream(void);
   /* Sets the internal stream to FILE, and saves the FILE's file
      descriptor in M_FD.  */
@@ -256,9 +334,9 @@ public:
 class tee_file : public ui_file
 {
 public:
-  /* Create a file which writes to both ONE and TWO.  ONE will remain
-     open when this object is destroyed; but TWO will be closed.  */
-  tee_file (ui_file *one, ui_file_up &&two);
+  /* Create a file which writes to both ONE and TWO.  Ownership of
+     both files is up to the user.  */
+  tee_file (ui_file *one, ui_file *two);
   ~tee_file () override;
 
   void write (const char *buf, long length_buf) override;
@@ -270,10 +348,28 @@ public:
   bool can_emit_style_escape () override;
   void flush () override;
 
+  void emit_style_escape (const ui_file_style &style) override
+  {
+    m_one->emit_style_escape (style);
+    m_two->emit_style_escape (style);
+  }
+
+  void reset_style () override
+  {
+    m_one->reset_style ();
+    m_two->reset_style ();
+  }
+
+  void puts_unfiltered (const char *str) override
+  {
+    m_one->puts_unfiltered (str);
+    m_two->puts_unfiltered (str);
+  }
+
 private:
   /* The two underlying ui_files.  */
   ui_file *m_one;
-  ui_file_up m_two;
+  ui_file *m_two;
 };
 
 /* A ui_file implementation that filters out terminal escape
@@ -290,6 +386,86 @@ public:
      sequences.  */
   void write (const char *buf, long length_buf) override;
   void puts (const char *linebuffer) override;
+
+  void emit_style_escape (const ui_file_style &style) override
+  {
+  }
+
+  void reset_style () override
+  {
+  }
 };
 
-#endif
+/* A base class for ui_file types that wrap another ui_file.  */
+
+class wrapped_file : public ui_file
+{
+public:
+
+  bool isatty () override
+  { return m_stream->isatty (); }
+
+  bool term_out () override
+  { return m_stream->term_out (); }
+
+  bool can_emit_style_escape () override
+  { return m_stream->can_emit_style_escape (); }
+
+  void flush () override
+  { m_stream->flush (); }
+
+  void wrap_here (int indent) override
+  { m_stream->wrap_here (indent); }
+
+  void emit_style_escape (const ui_file_style &style) override
+  { m_stream->emit_style_escape (style); }
+
+  /* Rest the current output style to the empty style.  */
+  void reset_style () override
+  { m_stream->reset_style (); }
+
+  int fd () const override
+  { return m_stream->fd (); }
+
+  void puts_unfiltered (const char *str) override
+  { m_stream->puts_unfiltered (str); }
+
+  void write_async_safe (const char *buf, long length_buf) override
+  { return m_stream->write_async_safe (buf, length_buf); }
+
+protected:
+
+  /* Note that this class does not assume ownership of the stream.
+     However, a subclass may choose to, by adding a 'delete' to its
+     destructor.  */
+  explicit wrapped_file (ui_file *stream)
+    : m_stream (stream)
+  {
+  }
+
+  /* The underlying stream.  */
+  ui_file *m_stream;
+};
+
+/* A ui_file that optionally puts a timestamp at the start of each
+   line of output.  */
+
+class timestamped_file : public wrapped_file
+{
+public:
+  explicit timestamped_file (ui_file *stream)
+    : wrapped_file (stream)
+  {
+  }
+
+  DISABLE_COPY_AND_ASSIGN (timestamped_file);
+
+  void write (const char *buf, long len) override;
+
+private:
+
+  /* True if the next output should be timestamped.  */
+  bool m_needs_timestamp = true;
+};
+
+#endif /* GDB_UI_FILE_H */
